from time import sleep

import pytest

from widgetastic_patternfly4 import BulletChart
from widgetastic_patternfly4.bulletchart import DataPoint
from widgetastic_patternfly4.bulletchart import Legend

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/charts/bullet-chart"

TEST_DATA = {
    "dot": {
        "id": "ws-react-c-chartbullet-bullet-chart-with-primary-measure-dot",
        "anchor": "#bullet-chart-with-primary-measure-dot",
        "bar_data": [
            DataPoint("Range", 75),
            DataPoint("Range", 50),
            DataPoint("Measure", 25),
            DataPoint("Measure", 60),
            DataPoint("Warning", 88),
        ],
        "legend_data": [
            Legend("Measure", 1),
            Legend("Measure", 2),
            Legend("Warning", None),
            Legend("Range", 1),
            Legend("Range", 2),
        ],
    },
    "tick": {
        "id": "ws-react-c-chartbullet-error-measure-and-custom-axis-ticks",
        "anchor": "#error-measure-and-custom-axis-ticks",
        "bar_data": [
            DataPoint("Range", 150),
            DataPoint("Range", 100),
            DataPoint("Range", 65),
            DataPoint("Measure", 75),
            DataPoint("Measure", 25),
            DataPoint("Error", 120),
            DataPoint("Warning", 80),
        ],
        "legend_data": [
            Legend("Measure", 1),
            Legend("Measure", 2),
            Legend("Warning", None),
            Legend("Error", None),
            Legend("Range", 1),
            Legend("Range", 2),
        ],
    },
}


@pytest.fixture(params=TEST_DATA.keys())
def chart_data(browser, request):
    sleep(3)  # Stabilized graph data on testing page; specially for firefox.
    # Firefox fails the test if the chart is not fully visible therefore we click here on anchor
    # in order to properly scroll down
    anchor = browser.element(f".//a[@href='{request.param['anchor']}']")
    browser.click(anchor)
    return (
        BulletChart(browser, id=TEST_DATA[request.param]["id"]),
        TEST_DATA[request.param]["bar_data"],
        TEST_DATA[request.param]["legend_data"],
        request.param,
    )


@pytest.mark.skip(reason="test is flaky")
def test_bullet_chart(chart_data):
    """Test BulletChart widget."""
    chart, bar_data, legend_data, chart_type = chart_data
    assert chart.is_displayed
    assert chart.legends == legend_data
    assert chart.data == bar_data
    assert "Warning" in chart.legend_names
    # get bar and check values
    warning_data = chart.get_data_point("Warning")
    expected_value = 88 if chart_type == "dot" else 80
    assert warning_data.value == expected_value
    assert warning_data.color == "rgb(236, 122, 8)"
    # get bar with wrong label
    assert not chart.get_data_point("foo")
    # get legend for check values
    warning_legend = chart.get_legend("Warning")
    assert warning_legend.value is None
    assert warning_legend.color == "rgb(236, 122, 8)"
    # get legend with wrong label
    assert not chart.get_legend("foo")



import math
from contextlib import contextmanager

from selenium.webdriver.common.keys import Keys
from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import GenericLocatorWidget
from widgetastic.widget import Text
from widgetastic.widget import TextInput
from widgetastic.widget import View

from .optionsmenu import OptionsMenu


class PaginationNavDisabled(Exception):
    pass


class BasePagination:
    """Represents the Patternfly pagination.

    https://www.patternfly.org/v4/documentation/react/components/pagination
    """

    DEFAULT_LOCATOR = (
        ".//div[contains(@class, 'pf-c-pagination') and not(contains(@class, 'pf-m-compact'))]"
    )

    _first = GenericLocatorWidget(".//button[contains(@data-action, 'first')]")
    _previous = GenericLocatorWidget(".//button[contains(@data-action, 'previous')]")
    _next = GenericLocatorWidget(".//button[contains(@data-action, 'next')]")
    _last = GenericLocatorWidget(".//button[contains(@data-action, 'last')]")
    _options = OptionsMenu()
    _items = Text(".//span[@class='pf-c-options-menu__toggle-text']")
    _current_page = TextInput(locator=".//input[@aria-label='Current page']")
    _total_pages = Text(".//div[@class='pf-c-pagination__nav-page-select']/span")

    @property
    def cached_per_page_value(self):
        return getattr(self, "_cached_per_page_value", None)

    @cached_per_page_value.setter
    def cached_per_page_value(self, value):
        self._cached_per_page_value = value

    @property
    def is_first_disabled(self):
        """Returns boolean detailing if the first page button is disabled."""
        return not self.browser.element(self._first).is_enabled()

    def first_page(self):
        """Clicks on the first page button."""
        if self.no_items or self.is_first_disabled:
            raise PaginationNavDisabled("first")
        self._first.click()

    @property
    def is_previous_disabled(self):
        """Returns boolean detailing if the previous page button is disabled."""
        return not self.browser.element(self._previous).is_enabled()

    def previous_page(self):
        """Clicks the previous page button."""
        if self.no_items or self.is_previous_disabled:
            raise PaginationNavDisabled("previous")
        self._previous.click()

    @property
    def is_next_disabled(self):
        """Returns boolean detailing if the next page button is disabled."""
        return not self.browser.element(self._next).is_enabled()

    def next_page(self):
        """Clicks the next page button."""
        if self.is_next_disabled:
            raise PaginationNavDisabled("next")
        self._next.click()

    @property
    def is_last_disabled(self):
        """Returns boolean detailing if the last page button is disabled."""
        return not self.browser.element(self._last).is_enabled()

    def last_page(self):
        """Clicks the last page button."""
        if self.is_last_disabled:
            raise PaginationNavDisabled("last")
        self._last.click()

    @property
    def current_page(self):
        """Returns an int of the current page number."""
        return int(self._current_page.value)

    @property
    def total_pages(self):
        """Returns int detailing the total number of pages."""
        return int(self._total_pages.text.strip().split()[1])

    @property
    def displayed_items(self):
        """Returns a string detailing the number of displayed items information.

        example "1 - 20 of 523 items"
        """
        items_string = self._items.text
        first_num, last_num = items_string.split("of")[0].split("-")
        return int(first_num.strip()), int(last_num.strip())

    @property
    def total_items(self):
        """Returns a string detailing the number of displayed items"""
        items_string = self._items.text
        return int(items_string.split("of")[1].split()[0])

    @property
    def per_page_options(self):
        """Returns an iterable of the available pagination options."""
        return self._options.items

    @property
    def no_items(self):
        """Returns wether the pagination object has elements or not"""
        return not self.total_items

    @property
    def current_per_page(self):
        """Returns an integer detailing how many items are shown per page."""
        if self.cached_per_page_value:
            return self.cached_per_page_value

        if self.no_items:
            return 0
        else:
            return int(self._options.selected_items[0].split()[0])

    @contextmanager
    def cache_per_page_value(self):
        """
        A context manager that can be used to prevent looking up the 'current page' value.

        This adds some efficiencies when iterating over pages or in cases where it is safe to
        assume that the "per page" setting is not going to change and it's not necessary to
        re-read it from the browser repeatedly.
        """
        self.cached_per_page_value = None
        self.cached_per_page_value = self.current_per_page
        yield
        self.cached_per_page_value = None

    def set_per_page(self, count):
        """Sets the number of items per page. (Will cast to str)"""
        value = str(count)
        value_per_page = "{} per page".format(value)
        items = self._options.items
        if value_per_page in items:
            self._options.item_select(value_per_page)
        elif value in items:
            self._options.item_select(value)
        else:
            raise ValueError(
                "count '{}' is not a valid option in the pagination dropdown".format(count)
            )

    def go_to_page(self, value):
        """Navigate to custom page number."""
        self._current_page.fill(value)
        self.browser.send_keys(Keys.RETURN, self._current_page)

    def __iter__(self):
        if self.current_page > 1:
            self.first_page()
        self._page_counter = 0
        return self

    def __next__(self):
        if self._page_counter < self.total_pages:
            self._page_counter += 1
            if self._page_counter > 1:
                self.next_page()
            return self._page_counter
        else:
            raise StopIteration


class Pagination(BasePagination, View):
    ROOT = ParametrizedLocator("{@locator}")

    def __init__(self, parent, locator=None, logger=None):
        super().__init__(parent=parent, logger=logger)
        if not locator:
            locator = self.DEFAULT_LOCATOR
        self.locator = locator


class BaseCompactPagination:
    @property
    def is_first_disabled(self):
        """Compact paginator has no 'first' button."""
        return self.is_previous_disabled

    def first_page(self):
        while not self.is_previous_disabled:
            self.previous_page()

    @property
    def is_last_disabled(self):
        """Compact paginator has no 'last' button."""
        return self.is_next_disabled

    def last_page(self):
        """Compact paginator has no "last" button, so iterates until last is reached."""
        while not self.is_next_disabled:
            self.next_page()

    @property
    def current_page(self):
        """
        Calculate the current page we are on.

        Compact pagination does not explicitly show this, so use some math.

        For example, if "per page" is set to '20', we know that a page displaying items:
            1-20 is on 20/20 = page 1
            21-40 is on page 40/20 = page 2
            41-60 is on page 60/20 = page 3

        and so on.
        """
        if self.no_items:
            return 0
        else:
            _, last_num = self.displayed_items
            return math.ceil(last_num / self.current_per_page)

    @property
    def total_pages(self):
        """
        Calculate total page count.

        Compact pagination does not explicitily show the page count, so use some math.
        """
        if self.no_items:
            return 0
        else:
            return math.ceil(self.total_items / self.current_per_page)


class CompactPagination(BaseCompactPagination, Pagination):
    DEFAULT_LOCATOR = (
        ".//div[contains(@class, 'pf-c-pagination') and contains(@class, 'pf-m-compact')]"
    )



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4 import Switch
from widgetastic_patternfly4 import SwitchDisabled

TESTING_PAGE_URL = "https://patternfly-react.surge.sh/components/switch"


@pytest.fixture
def view(browser):
    class TestView(View):
        switch = Switch(locator='.//label[@for="simple-switch"]')
        no_label_switch = Switch(locator='.//label[@for="no-label-switch-on"]')
        disabled_switch_on = Switch(locator='.//label[@for="disabled-switch-on"]')
        disabled_switch_off = Switch(locator='.//label[@for="disabled-switch-off"]')
        disabled_no_label_switch_on = Switch(locator='.//label[@for="disabled-no-label-switch-on"]')
        disabled_no_label_switch_off = Switch(
            locator='.//label[@for="disabled-no-label-switch-off"]'
        )

    return TestView(browser)


def test_switch_is_displayed(view):
    assert view.switch.is_displayed
    assert view.no_label_switch.is_displayed
    assert view.disabled_switch_on.is_displayed
    assert view.disabled_switch_off.is_displayed
    assert view.disabled_no_label_switch_on.is_displayed
    assert view.disabled_no_label_switch_off.is_displayed


def test_switch_is_enabled(view):
    assert view.switch.is_enabled
    assert view.no_label_switch.is_enabled
    assert not view.disabled_switch_on.is_enabled
    assert not view.disabled_switch_off.is_enabled
    assert not view.disabled_no_label_switch_on.is_enabled
    assert not view.disabled_no_label_switch_off.is_enabled


def test_switch_label(view):
    assert view.switch.label == "Message when on"
    assert view.no_label_switch.label is None
    assert view.disabled_switch_on.label == "Message when on"
    assert view.disabled_switch_off.label == "Message when off"
    assert view.disabled_no_label_switch_on.label is None
    assert view.disabled_no_label_switch_off.label is None


def test_switch_selected(view):
    assert view.switch.read()
    assert view.no_label_switch.read()
    assert view.disabled_switch_on.read()
    assert not view.disabled_switch_off.read()
    assert view.disabled_no_label_switch_on.read()
    assert not view.disabled_no_label_switch_off.read()


def test_switch_fill(view):
    assert view.switch.selected
    assert view.switch.label == "Message when on"
    assert not view.switch.fill(True)
    assert view.switch.selected
    assert view.switch.label == "Message when on"
    assert view.switch.fill(False)
    assert not view.switch.selected
    assert view.switch.label == "Message when off"
    assert view.switch.fill(True)
    assert view.switch.selected
    assert view.switch.label == "Message when on"


def test_switch_fill_disabled(view):
    for disabled_switch in (
        view.disabled_switch_on,
        view.disabled_switch_off,
        view.disabled_no_label_switch_on,
        view.disabled_no_label_switch_off,
    ):
        with pytest.raises(SwitchDisabled):
            disabled_switch.fill(True)



import pytest
from widgetastic.widget import View

from widgetastic_patternfly4.ouia import Button

TESTING_PAGE_URL = "https://patternfly-docs-ouia.netlify.app/documentation/react/components/button"


@pytest.fixture
def view(browser):
    class TestView(View):
        ROOT = ".//div[@id='ws-react-c-button-ouia']"
        button = Button("Primary")

    return TestView(browser)


def test_button_click(view):
    view.button.__repr__()
    assert view.button.is_displayed
    view.button.click()



from widgetastic.ouia import OUIAGenericView
from widgetastic.ouia import OUIAGenericWidget
from widgetastic.widget.table import Table
from widgetastic.xpath import quote

from widgetastic_patternfly4.alert import BaseAlert
from widgetastic_patternfly4.breadcrumb import BaseBreadCrumb
from widgetastic_patternfly4.button import BaseButton
from widgetastic_patternfly4.card import BaseCard
from widgetastic_patternfly4.contextselector import BaseContextSelector
from widgetastic_patternfly4.dropdown import BaseDropdown
from widgetastic_patternfly4.dropdown import BaseGroupDropdown
from widgetastic_patternfly4.dropdown import BaseSplitButtonDropdown
from widgetastic_patternfly4.formselect import BaseFormSelect
from widgetastic_patternfly4.menu import BaseCheckboxMenu
from widgetastic_patternfly4.menu import BaseMenu
from widgetastic_patternfly4.modal import BaseModal
from widgetastic_patternfly4.navigation import BaseNavigation
from widgetastic_patternfly4.optionsmenu import BaseOptionsMenu
from widgetastic_patternfly4.pagination import BaseCompactPagination
from widgetastic_patternfly4.pagination import BasePagination
from widgetastic_patternfly4.select import BaseCheckboxSelect
from widgetastic_patternfly4.select import BaseSelect
from widgetastic_patternfly4.switch import BaseSwitch
from widgetastic_patternfly4.table import BaseExpandableTable
from widgetastic_patternfly4.table import BasePatternflyTable


class Alert(BaseAlert, OUIAGenericWidget):
    OUIA_COMPONENT_TYPE = "PF4/Alert"


class BreadCrumb(BaseBreadCrumb, OUIAGenericWidget):
    OUIA_COMPONENT_TYPE = "PF4/Breadcrumb"


class Button(BaseButton, OUIAGenericWidget):
    OUIA_COMPONENT_TYPE = "PF4/Button"


class Card(BaseCard, OUIAGenericWidget):
    OUIA_COMPONENT_TYPE = "PF4/Card"


class Dropdown(BaseDropdown, OUIAGenericWidget):
    OUIA_COMPONENT_TYPE = "PF4/Dropdown"


class GroupDropdown(BaseGroupDropdown, Dropdown):
    pass


class SplitButtonDropdown(BaseSplitButtonDropdown, Dropdown):
    pass


class FormSelect(BaseFormSelect, OUIAGenericWidget):
    OUIA_COMPONENT_TYPE = "PF4/FormSelect"


class Menu(BaseMenu, Dropdown):
    OUIA_COMPONENT_TYPE = "PF4/Menu"


class CheckboxMenu(BaseCheckboxMenu, Dropdown):
    OUIA_COMPONENT_TYPE = "PF4/Menu"


class Modal(BaseModal, OUIAGenericView):
    OUIA_COMPONENT_TYPE = "PF4/ModalContent"


class Navigation(BaseNavigation, OUIAGenericWidget):
    OUIA_COMPONENT_TYPE = "PF4/Nav"


class Pagination(BasePagination, OUIAGenericView):
    OUIA_COMPONENT_TYPE = "PF4/Pagination"


class CompactPagination(BaseCompactPagination, Pagination):
    pass


class CheckboxSelect(BaseCheckboxSelect, Dropdown):
    OUIA_COMPONENT_TYPE = "PF4/Select"


class Switch(BaseSwitch, OUIAGenericWidget):
    OUIA_COMPONENT_TYPE = "PF4/Switch"


class PatternflyTable(BasePatternflyTable, Table):
    def __init__(
        self,
        parent,
        component_id,
        column_widgets=None,
        assoc_column=None,
        rows_ignore_top=None,
        rows_ignore_bottom=None,
        top_ignore_fill=False,
        bottom_ignore_fill=False,
        logger=None,
    ):
        self.component_type = "PF4/Table"
        self.component_id = component_id
        super().__init__(
            parent,
            locator=(
                f".//*[@data-ouia-component-type={quote(self.component_type)} "
                f"and @data-ouia-component-id={quote(self.component_id)}]"
            ),
            column_widgets=column_widgets,
            assoc_column=assoc_column,
            rows_ignore_top=rows_ignore_top,
            rows_ignore_bottom=rows_ignore_bottom,
            top_ignore_fill=top_ignore_fill,
            bottom_ignore_fill=bottom_ignore_fill,
            logger=logger,
        )


class ExpandableTable(BaseExpandableTable, PatternflyTable):
    pass


class Select(BaseSelect, Dropdown):
    OUIA_COMPONENT_TYPE = "PF4/Select"


class ContextSelector(BaseContextSelector, Select):
    OUIA_COMPONENT_TYPE = "PF4/ContextSelector"


class OptionsMenu(BaseOptionsMenu, Dropdown):
    OUIA_COMPONENT_TYPE = "PF4/OptionsMenu"



from widgetastic.utils import ParametrizedLocator
from widgetastic.widget import Widget


class BaseAlert:
    """Represents alert block.

    https://www.patternfly.org/v4/documentation/react/components/alert
    """

    TITLE = './/h4[@class="pf-c-alert__title"]'
    DESCRIPTION = './/div[@class="pf-c-alert__description"]'
    ACTION = './/div[contains(@class, "pf-c-alert__action")]'
    TYPE_MAPPING = {
        "pf-m-warning": "warning",
        "pf-m-success": "success",
        "pf-m-danger": "error",
        "pf-m-info": "info",
    }

    def read(self):
        """Returns the text of the body of the alert as a string."""
        return self.body

    @property
    def _raw_title_el(self):
        return self.browser.element(self.TITLE)

    @property
    def _raw_description_el(self):
        return self.browser.element(self.DESCRIPTION)

    @property
    def title(self):
        """Returns the title text of the alert as a string."""
        trim_text = self.browser.text(self.browser.element("./span", parent=self._raw_title_el))
        return self.browser.text(self._raw_title_el)[len(trim_text) :].strip()

    @property
    def body(self):
        """Returns the text of the body of the alert as a string."""
        el = self.browser.element(self.DESCRIPTION)
        return self.browser.text(el)

    def click_action(self):
        """Clicks the defined action button of the alert."""
        el = self.browser.element(self.ACTION)
        self.browser.click(el)

    def click_link(self):
        el = self.browser.element("./a", parent=self._raw_description_el)
        self.browser.click(el)

    @property
    def type(self):
        """Returns the type of the alert, one of warning, success, error or info."""
        for class_ in self.browser.classes(self):
            if class_ in self.TYPE_MAPPING:
                return self.TYPE_MAPPING[class_]
        else:
            raise ValueError(
                "Could not find a proper alert type."
                f"\nAvailable classes: {self.TYPE_MAPPING!r} "
                f"\nAlert has: {self.browser.classes(self)!r}"
            )

    def assert_no_error(self):
        """Asserts that the warning is not of the error type."""
        if self.type == "error":
            raise AssertionError(f"assert_no_error: {self.body}")

    def __repr__(self):
        return f"{type(self).__name__}({self.locator!r})"


class Alert(BaseAlert, Widget):
    ROOT = ParametrizedLocator("{@locator}")

    def __init__(self, parent, locator, logger=None):
        super().__init__(parent, logger=logger)
        self.locator = locator
